theory Abstract_Completion_Old
imports Abstract_Completion
begin

locale incfam =
  fixes r :: "nat \<Rightarrow> 'a \<Rightarrow> 'b rel"
    and less :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<succ>" 50)
  assumes SN_less: "SN {(x, y). x \<succ> y}"
    and inc: "a \<succ> b \<Longrightarrow> r i b \<subseteq> r i a"
begin

definition "r_inf a = (\<Union>i. r i a)"
definition "r_per a = (\<Union>i. \<Inter>j\<in>{j. j \<ge> i}. r j a)"
definition "r_bel i a = (\<Union>b\<in>{b. a \<succ> b}. r i b)"

lemma r_bel_subset:
  "r_bel i a \<subseteq> r i a"
using inc by (auto simp: r_bel_def)

lemma r_per_inc:
  "a \<succ> b \<Longrightarrow> r_per b \<subseteq> r_per a"
using inc by (force simp: r_per_def)

(*Theorem 3*)
lemma r_inf_r_per_conv:
  assumes "\<And>i a. r i a \<subseteq> r (Suc i) a \<union> (r_bel (Suc i) a)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "r_inf a \<subseteq> (r_per a)\<^sup>\<leftrightarrow>\<^sup>*"
proof (induct rule: SN_induct [OF SN_less])
  case (1 a)
  then have IH: "\<And>i. r_bel i a \<subseteq> (r_per a)\<^sup>\<leftrightarrow>\<^sup>*"
    by (auto simp: r_inf_def r_bel_def dest: r_per_inc [THEN conversion_mono])
    
  show ?case
  proof
    fix x y assume "(x, y) \<in> r_inf a"
    then obtain i where *: "(x, y) \<in> r i a" by (auto simp: r_inf_def)
    show "(x, y) \<in> (r_per a)\<^sup>\<leftrightarrow>\<^sup>*"
    proof (cases "(x, y) \<in> r_per a")
      case False
      with * have "\<exists>j>i. (x, y) \<notin> r j a"
        using Suc_le_eq unfolding r_per_def by blast
      moreover def j \<equiv> "LEAST j. j > i \<and> (x, y) \<notin> r j a"
      ultimately have "j > i" and **: "(x, y) \<notin> r j a" by (metis (lifting) LeastI)+
      then have "j - 1 < j" and [simp]: "Suc (j - Suc 0) = j" by auto
      from not_less_Least [OF this(1) [unfolded j_def], folded j_def] and * and \<open>j > i\<close>
        have "(x, y) \<in> r (j - 1) a" by (cases "j = Suc i") auto
      with assms [of "j - 1" a] and ** have "(x, y) \<in> (r_bel j a)\<^sup>\<leftrightarrow>\<^sup>*" by auto
      with conversion_mono [OF IH, of j] show ?thesis by auto
    qed auto
  qed
qed

end

context kb
begin

abbreviation "lless \<equiv> s_mul_ext {\<cdot>\<unrhd>} ((relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+)"
abbreviation "lleq \<equiv> ns_mul_ext {\<cdot>\<unrhd>} ((relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+)"

(*order pair for labels*)
lemma lop: "SN_order_pair lless lleq"
proof (intro SN_order_pair.mul_ext_SN_order_pair)
  show "SN_order_pair ((relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+) {\<cdot>\<unrhd>}"
  proof (standard)
    show "{\<cdot>\<unrhd>} O (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+ \<subseteq> (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+" by regexp
    show "(relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+ O {\<cdot>\<unrhd>}  \<subseteq> (relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+" by regexp
    show "SN ((relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+)"
      unfolding SN_trancl_SN_conv by (blast intro: SN_encomp_Un_less_relto_encompeq)
  qed (auto simp: refl_on_def trans_def intro: encompeq_trans)
qed

interpretation lop: SN_order_pair lless lleq by (rule lop)

definition "lstep E R \<alpha> = {(s, t).
  (\<alpha>, {#s, s, s, t, t, t#}) \<in> lleq \<and> (s, t) \<in> rstep E \<or>
  (\<exists>(l, r)\<in>R. (\<alpha>, {#s, l, r#}) \<in> lleq \<and> (s, t) \<in> rstep_rule (l, r))}"

lemma lstep_cases [consumes 1, case_names eq rule]:
  assumes "(s, t) \<in> lstep E R \<alpha>"
  obtains "(\<alpha>, {#s, s, s, t, t, t#}) \<in> lleq" and "(s, t) \<in> rstep E"
  | l r where "(l, r) \<in> R" and "(\<alpha>, {#s, l , r#}) \<in> lleq" and "(s, t) \<in> rstep_rule (l, r)"
using assms by (auto simp: lstep_def)

lemma lstep_E_I:
  assumes "(\<alpha>, {#s, s, s, t, t, t#}) \<in> lleq" and "(s, t) \<in> rstep E"
  shows "(s, t) \<in> lstep E R \<alpha>"
using assms by (auto simp: lstep_def)

lemma lstep_R_I:
  assumes "(\<alpha>, {#s, l, r#}) \<in> lleq" and "(l, r) \<in> R" and "(s, t) \<in> rstep_rule (l, r)"
  shows "(s, t) \<in> lstep E R \<alpha>"
using assms by (auto simp: lstep_def)

lemma lstep_inc:
  assumes "(\<alpha>, \<beta>) \<in> lless"
  shows "lstep E R \<beta> \<subseteq> lstep E R \<alpha>"
proof
  fix s t assume "(s, t) \<in> lstep E R \<beta>"
  then show "(s, t) \<in> lstep E R \<alpha>"
  proof (cases rule: lstep_cases)
    case (eq)
    with assms have "(\<alpha>, {#s, s, s, t, t, t#}) \<in> lleq"
      by (blast dest: lop.trans_NS_point s_ns_mul_ext)
    with eq show ?thesis by (auto simp: lstep_def)
  next
    case (rule l r)
    with assms have "(\<alpha>, {#s, l, r#}) \<in> lleq"
      by (blast dest: lop.trans_NS_point s_ns_mul_ext)
    with rule show ?thesis unfolding lstep_def by blast
  qed
qed

context
  fixes R and R'
  assumes less: "R \<union> R' \<subseteq> {\<succ>}"
begin

lemma
  rstep_R: "rstep R \<subseteq> {\<succ>}" and
  rstep_R': "rstep R' \<subseteq> {\<succ>}"
using less by (blast intro: ctxt subst)+

lemma simplify:
  assumes \<alpha>: "(\<alpha>, {#s, s, s, u, u, u#}) \<in> lleq"
   and st: "(s, t) \<in> rstep R'" and tu: "(t, u) \<in> ((rstep E')\<^sup>\<leftrightarrow>)\<^sup>="
  shows "(s, u) \<in> (\<Union>\<beta>\<in>{\<beta>. (\<alpha>, \<beta>) \<in> lless}. lstep E' R' \<beta>)\<^sup>\<leftrightarrow>\<^sup>*" (is "_ \<in> ?R")
proof -
  let ?S = "(relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+"
  have "s \<succ> t" using st and rstep_R' by blast
  moreover def \<gamma> \<equiv> "{#t, t, t, u, u, u#}"
  ultimately have "({#s, s, s, u, u, u#}, \<gamma>) \<in> lless"
    by (intro s_mul_extI [of _ "[u, u, u]" "{#s, s, s#}" _ "[u, u, u]" "{#t, t, t#}"])
       (auto simp: ac_simps split: if_splits)
  with \<alpha> have "(\<alpha>, \<gamma>) \<in> lless" using lop.compat_NS_S_point by blast
  moreover have "(t, u) \<in> ((lstep E' R' \<gamma>)\<^sup>\<leftrightarrow>)\<^sup>="
     using tu and lop.refl_NS_point by (auto simp: \<gamma>_def ac_simps intro!: lstep_E_I)
  ultimately have "(t, u) \<in> ?R" by blast
  moreover have "(s, t) \<in> ?R"
  proof -
    obtain C \<sigma> l r where "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and "(l, r) \<in> R'" using st by blast
    then have st: "(s, t) \<in> rstep_rule (l, r)" and "s \<cdot>\<unrhd> l" and "l \<succ> r" using rstep_R' by auto
    moreover then have "(s, r) \<in> ?S" by blast
    moreover def \<beta> \<equiv> "{#s, l, r#}"
    ultimately have "({#s, s, s, u, u, u#}, \<beta>) \<in> lless"
      by (intro s_mul_extI [of _ "[s, s]" "{#s, u, u, u#}" _ "[s, l]" "{#r#}"])
         (auto simp: ac_simps less_Suc_eq split: if_splits)
    with \<alpha> have "(\<alpha>, \<beta>) \<in> lless" using lop.compat_NS_S_point by blast
    moreover have "(s, t) \<in> lstep E' R' \<beta>"
      using st and \<open>(l, r) \<in> R'\<close> and lop.refl_NS_point by (auto intro!: lstep_R_I simp: \<beta>_def)
    ultimately show ?thesis by blast
  qed
  ultimately show "(s, u) \<in> ?R" by (auto simp: conversion_def)
qed

context
  fixes E E'
  assumes KB: "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')"
begin

lemma KB_lstep_E_subset:
  fixes s t
  defines "\<alpha> \<equiv> {#s, s, s, t, t, t#}"
  assumes eq: "(s, t) \<in> E"
  shows "(s, t) \<in> E' \<or> (s, t) \<in> (\<Union>\<beta>\<in>{\<beta>. (\<alpha>, \<beta>) \<in> lless}. lstep E' R' \<beta>)\<^sup>\<leftrightarrow>\<^sup>*"
    (is "_ \<or> _ \<in> ?R")
proof -
  have \<alpha>: "(\<alpha>, {#s, s, s, t, t, t#}) \<in> lleq"
    using lop.refl_NS_point by (auto simp: \<alpha>_def)

  consider "(s, t) \<in> E'" | "(s, t) \<in> (rstep R' O (rstep E')\<^sup>=)"
    | "(s, t) \<in> (rstep E')\<^sup>= O (rstep R')\<inverse>" | "s = t"
    using eq and KB_E_subset [OF KB] by blast
  then show ?thesis
  proof (cases)
    case 2
    then obtain u where "(s, u) \<in> rstep R'" and "(u, t) \<in> (rstep E')\<^sup>=" by blast
    then show ?thesis using \<alpha> by (blast intro: simplify)
  next
    case 3
    then obtain u where "(t, u) \<in> rstep R'" and "(u, s) \<in> ((rstep E')\<^sup>\<leftrightarrow>)\<^sup>=" by blast
    from simplify [OF _ this] have "(s, t) \<in> ?R" using \<alpha> by (simp add: conversion_inv ac_simps)
    then show ?thesis by blast
  qed auto
qed

context
  assumes "enc"
begin

lemma KB_lstep_R_subset:
  fixes s t
  defines "\<alpha> \<equiv> {#s, s, t#}"
  assumes rule: "(s, t) \<in> R"
  shows "(s, t) \<in> R' \<or> (s, t) \<in> (\<Union>\<beta>\<in>{\<beta>. (\<alpha>, \<beta>) \<in> lless}. lstep E' R' \<beta>)\<^sup>\<leftrightarrow>\<^sup>*"
    (is "_ \<or> _ \<in> ?R")
proof -
  let ?S = "(relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+"
  have \<alpha>: "(\<alpha>, {#s, s, t#}) \<in> lleq" using lop.refl_NS_point by (auto simp: \<alpha>_def)
  have "s \<succ> t" using rule and rstep_R by blast
  consider "(s, t) \<in> R'" | "(s, t) \<in> rstep_enc R' O rstep E'" | "(s, t) \<in> rstep R' O (rstep R')\<inverse>"
    using rule and KB_R_subset [OF KB] and rule by blast
  then show ?thesis
  proof (cases)
    case 2
    then obtain u l r where "(l, r) \<in> R'" and lu: "(s, u) \<in> rstep_rule (l, r)"
      and "s \<cdot>\<rhd> l" and ut: "(u, t) \<in> rstep E'"
      by (auto simp: rstep_enc_def split: if_splits intro: \<open>enc\<close>)
    from rstep_rule_imp_rstep [OF this(1,2)]
      have "(s, u) \<in> rstep R'" by auto
    then have "s \<succ> u" using rstep_R' by blast
    moreover def \<gamma> \<equiv> "{#t, t, t, u, u, u#}"
    ultimately have "({#s, s, t#}, \<gamma>) \<in> lless"
      using \<open>s \<succ> t\<close>
      by (intro s_mul_extI [of _ "[]" "{#s, s, t#}" _ "[]" \<gamma>])
         (force split: if_splits intro!: exI [of _ s])+
    then have "(\<alpha>, \<gamma>) \<in> lless" using \<alpha> and lop.compat_NS_S_point by blast
    moreover have "(u, t) \<in> lstep E' R' \<gamma>"
      using lop.refl_NS_point and ut
      by (intro lstep_E_I) (auto simp: \<gamma>_def ac_simps)
    ultimately have "(u, t) \<in> ?R" by blast
    moreover have "(s, u) \<in> ?R"
    proof -
      def \<beta> \<equiv> "{#s, l, r#}"
      moreover have "(s, r) \<in> ?S"
      proof -
        have "(s, l) \<in> ?S" using \<open>s \<cdot>\<rhd> l\<close> by blast
        moreover have "(l, r) \<in> ?S" using \<open>(l, r) \<in> R'\<close> and less by blast
        ultimately show ?thesis by auto
      qed
      ultimately have "({#s, s, t#}, \<beta>) \<in> lless"
        using \<open>s \<cdot>\<rhd> l\<close>
        by (intro s_mul_extI [of _ "[s]" "{#s, t#}" _ "[s]" "{#l, r#}"])
           (auto split: if_splits)
      with rule have "(\<alpha>, \<beta>) \<in> lless" using \<alpha> and lop.compat_NS_S_point by blast
      moreover have "(s, u) \<in> lstep E' R' \<beta>"
        using lop.refl_NS_point and lu
        by (intro lstep_R_I [OF _ \<open>(l, r) \<in> R'\<close>]) (auto simp: \<beta>_def)
      ultimately show ?thesis by blast
    qed
    ultimately have "(s, t) \<in> ?R" by (auto simp: conversion_def)
    then show ?thesis by blast
  next
    case 3
    then obtain u where "(s, u) \<in> rstep R'" and tu: "(t, u) \<in> rstep R'" by blast
    then obtain l\<^sub>1 r\<^sub>1 l\<^sub>2 r\<^sub>2
      where rule\<^sub>1: "(l\<^sub>1, r\<^sub>1) \<in> R'" and "(s, u) \<in> rstep_rule (l\<^sub>1, r\<^sub>1)" and "s \<cdot>\<unrhd> l\<^sub>1" and "u \<cdot>\<unrhd> r\<^sub>1"
      and rule\<^sub>2: "(l\<^sub>2, r\<^sub>2) \<in> R'" and "(t, u) \<in> rstep_rule (l\<^sub>2, r\<^sub>2)" and "t \<cdot>\<unrhd> l\<^sub>2" and "u \<cdot>\<unrhd> r\<^sub>2"
      by (blast elim: rstep_imp_rstep_rule)
    moreover have "t \<succ> u" and "s \<succ> t"
      using tu and rule and rstep_R and rstep_R' by (auto dest: rstep_rule_imp_rstep)
    ultimately have "(s, u) \<in> lstep E' R' {#s, l\<^sub>1, r\<^sub>1#}" (is "_ \<in> lstep _ _ ?\<beta>")
      and "(t, u) \<in> lstep E' R' {#t, l\<^sub>2, r\<^sub>2#}" (is "_ \<in> lstep _ _ ?\<gamma>")
      using rule\<^sub>1 and rule\<^sub>2 by (auto intro!: lstep_R_I lop.refl_NS_point)
    moreover have "(\<alpha>, ?\<beta>) \<in> lless"
    proof -
      have "(t, r\<^sub>1) \<in> ?S" using \<open>t \<succ> u\<close> and \<open>u \<cdot>\<unrhd> r\<^sub>1\<close> by blast
      then have "({#s, s, t#}, ?\<beta>) \<in> lless"
        using \<open>s \<cdot>\<unrhd> l\<^sub>1\<close>
        by (intro s_mul_extI [of _ "[s, s]" "{#t#}" _ "[s, l\<^sub>1]" "{#r\<^sub>1#}"])
           (auto simp: ac_simps less_Suc_eq split: if_splits)
      with \<alpha> show ?thesis
        using lop.compat_NS_S_point by blast
    qed
    moreover have "(\<alpha>, ?\<gamma>) \<in> lless"
    proof -
      have "(s, l\<^sub>2) \<in> ?S" using \<open>s \<succ> t\<close> and \<open>t \<cdot>\<unrhd> l\<^sub>2\<close> by blast
      moreover have "(t, r\<^sub>2) \<in> ?S" using \<open>t \<succ> u\<close> and \<open>u \<cdot>\<unrhd> r\<^sub>2\<close> by blast
      moreover have "(s, t) \<in> ?S" using \<open>s \<succ> t\<close> by blast
      ultimately have "({#s, s, t#}, ?\<gamma>) \<in> lless"
        by (intro s_mul_extI [of _ "[]" "{#s, s, t#}" _ "[]" "{#t, l\<^sub>2, r\<^sub>2#}"])
           (auto split: if_splits)
      with \<alpha> show ?thesis
        using lop.compat_NS_S_point by blast
    qed
    ultimately have "(s, u) \<in> ?R" and "(u, t) \<in> ?R" by blast+
    then have "(s, t) \<in> ?R" by (auto simp: conversion_def)
    then show ?thesis by blast
  qed auto
qed

lemma KB_lstep_subset:
  "lstep E R \<alpha> \<subseteq> lstep E' R' \<alpha> \<union> (\<Union>\<beta>\<in>{\<beta>. (\<alpha>, \<beta>) \<in> lless}. lstep E' R' \<beta>)\<^sup>\<leftrightarrow>\<^sup>*"
    (is "_ \<subseteq> lstep E' R' \<alpha> \<union> ?R")
proof
  let ?S = "(relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+"

  fix s t assume lstep: "(s, t) \<in> lstep E R \<alpha>"
  then show "(s, t) \<in> lstep E' R' \<alpha> \<union> ?R"
  proof (cases rule: lstep_cases)
    case eq
    then obtain C \<sigma> l r where "(l, r) \<in> E" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by blast
    then consider "(l, r) \<in> E'" | "(l, r) \<in> (rstep R' O (rstep E')\<^sup>=)"
      | "(l, r) \<in> (rstep E')\<^sup>= O (rstep R')\<inverse>" | "l = r"
      using KB_E_subset [OF KB] by blast
    then show ?thesis
    proof (cases)
      case 1
      then have "(s, t) \<in> rstep E'" by (auto simp: s t)
      with eq(1) have "(s, t) \<in> lstep E' R' \<alpha>" by (intro lstep_E_I)
      then show ?thesis by blast
    next
      case 2
      then obtain u where "(s, u) \<in> rstep R'" and "(u, t) \<in> (rstep E')\<^sup>=" unfolding s t by blast
      then show ?thesis using eq by (blast intro: simplify)
    next
      case 3
      then obtain u where "(t, u) \<in> rstep R'" and "(u, s) \<in> ((rstep E')\<^sup>\<leftrightarrow>)\<^sup>=" unfolding s t by blast
      from simplify [OF _ this] have "(s, t) \<in> ?R" using eq by (simp add: conversion_inv ac_simps)
      then show ?thesis by blast
    qed (auto simp: s t)
  next
    case (rule l r)
    then obtain C \<sigma> where s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by (auto elim: rstep_rule.cases)
    with less and rule have "l \<succ> r" and "s \<succ> t" by (auto intro: subst ctxt)

    consider "(l, r) \<in> R'" | "(l, r) \<in> rstep_enc R' O rstep E'" | "(l, r) \<in> rstep R' O (rstep R')\<inverse>"
      using KB_R_subset [OF KB] and rule by blast
    then show ?thesis
    proof (cases)
      case 1
      with rule have "(s, t) \<in> lstep E' R' \<alpha>" by (intro lstep_R_I)
      then show ?thesis by blast
    next
      case 2
      then obtain u l' r' where "(l', r') \<in> R'" and lu: "(l, u) \<in> rstep_rule (l', r')"
        and "l \<cdot>\<rhd> l'" and ur: "(u, r) \<in> rstep E'"
        by (auto simp: rstep_enc_def split: if_splits intro: \<open>enc\<close>)
      from rstep_rule_imp_rstep [OF this(1,2)]
        have "(s, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> rstep R'" (is "(_, ?u) \<in> rstep R'") by (auto simp: s)
      then have "s \<succ> ?u" using rstep_R' by blast
      moreover def \<gamma> \<equiv> "{#t, t, t, ?u, ?u, ?u#}"
      ultimately have "({#s, l, r#}, \<gamma>) \<in> lless"
        using \<open>s \<succ> t\<close>
        by (intro s_mul_extI [of _ "[]" "{#s, l, r#}" _ "[]" \<gamma>])
           (force split: if_splits intro!: exI [of _ s])+
      then have "(\<alpha>, \<gamma>) \<in> lless" using rule and lop.compat_NS_S_point by blast
      moreover have "(?u, t) \<in> lstep E' R' \<gamma>"
        using lop.refl_NS_point and ur
        by (intro lstep_E_I) (auto simp: \<gamma>_def t ac_simps)
      ultimately have "(?u, t) \<in> ?R" by blast
      moreover have "(s, ?u) \<in> ?R"
      proof -
        def \<beta> \<equiv> "{#s, l', r'#}"
        moreover have "(l, r') \<in> ?S"
        proof -
          have "(l, l') \<in> ?S" using \<open>l \<cdot>\<rhd> l'\<close> by blast
          moreover have "(l', r') \<in> ?S" using \<open>(l', r') \<in> R'\<close> and less by blast
          ultimately show ?thesis by auto
        qed
        ultimately have "({#s, l, r#}, \<beta>) \<in> lless"
          using \<open>l \<cdot>\<rhd> l'\<close>
          by (intro s_mul_extI [of _ "[s]" "{#l, r#}" _ "[s]" "{#l', r'#}"])
             (auto split: if_splits)
        with rule have "(\<alpha>, \<beta>) \<in> lless" using lop.compat_NS_S_point by blast
        moreover have "(s, ?u) \<in> lstep E' R' \<beta>"
          using lop.refl_NS_point and lu
          by (intro lstep_R_I [OF _ \<open>(l', r') \<in> R'\<close>])
             (auto simp: \<beta>_def s intro: rstep_rule_ctxt rstep_rule_subst)
        ultimately show ?thesis by blast
      qed
      ultimately have "(s, t) \<in> ?R" by (auto simp: conversion_def)
      then show ?thesis by blast
    next
      case 3
      then obtain u where "(l, u) \<in> rstep R'" and ru: "(r, u) \<in> rstep R'" by blast
      then obtain l\<^sub>1 r\<^sub>1 l\<^sub>2 r\<^sub>2
        where rule\<^sub>1: "(l\<^sub>1, r\<^sub>1) \<in> R'" and "(l, u) \<in> rstep_rule (l\<^sub>1, r\<^sub>1)" and "l \<cdot>\<unrhd> l\<^sub>1" and "u \<cdot>\<unrhd> r\<^sub>1"
        and rule\<^sub>2: "(l\<^sub>2, r\<^sub>2) \<in> R'" and "(r, u) \<in> rstep_rule (l\<^sub>2, r\<^sub>2)" and "r \<cdot>\<unrhd> l\<^sub>2" and "u \<cdot>\<unrhd> r\<^sub>2"
        by (blast elim: rstep_imp_rstep_rule)
      moreover have "r \<succ> u" and "s \<succ> t"
        using ru and rule and rstep_R and rstep_R' by (auto dest: rstep_rule_imp_rstep)
      ultimately have "(s, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> rstep_rule (l\<^sub>1, r\<^sub>1)"
        and "(t, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> rstep_rule (l\<^sub>2, r\<^sub>2)"
        by (auto simp: s t intro: rstep_rule_ctxt rstep_rule_subst)
      then have "(s, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> lstep E' R' {#s, l\<^sub>1, r\<^sub>1#}" (is "_ \<in> lstep _ _ ?\<beta>")
        and "(t, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> lstep E' R' {#t, l\<^sub>2, r\<^sub>2#}" (is "_ \<in> lstep _ _ ?\<gamma>")
        using rule\<^sub>1 and rule\<^sub>2 by (auto intro!: lstep_R_I lop.refl_NS_point)
      moreover have "(\<alpha>, ?\<beta>) \<in> lless"
      proof -
        have "(r, r\<^sub>1) \<in> ?S" using \<open>r \<succ> u\<close> and \<open>u \<cdot>\<unrhd> r\<^sub>1\<close> by blast
        then have "({#s, l, r#}, ?\<beta>) \<in> lless"
          using \<open>l \<cdot>\<unrhd> l\<^sub>1\<close>
          by (intro s_mul_extI [of _ "[s, l]" "{#r#}" _ "[s, l\<^sub>1]" "{#r\<^sub>1#}"])
             (auto simp: ac_simps less_Suc_eq split: if_splits)
        with rule(2) show ?thesis
          using lop.compat_NS_S_point by blast
      qed
      moreover have "(\<alpha>, ?\<gamma>) \<in> lless"
      proof -
        have "(l, l\<^sub>2) \<in> ?S" using \<open>l \<succ> r\<close> and \<open>r \<cdot>\<unrhd> l\<^sub>2\<close> by blast
        moreover have "(r, r\<^sub>2) \<in> ?S" using \<open>r \<succ> u\<close> and \<open>u \<cdot>\<unrhd> r\<^sub>2\<close> by blast
        moreover have "(s, t) \<in> ?S" using \<open>s \<succ> t\<close> by blast
        ultimately have "({#s, l, r#}, ?\<gamma>) \<in> lless"
          by (intro s_mul_extI [of _ "[]" "{#s, l, r#}" _ "[]" "{#t, l\<^sub>2, r\<^sub>2#}"])
             (auto split: if_splits)
        with rule(2) show ?thesis
          using lop.compat_NS_S_point by blast
      qed
      ultimately have "(s, C\<langle>u \<cdot> \<sigma>\<rangle>) \<in> ?R" and "(C\<langle>u \<cdot> \<sigma>\<rangle>, t) \<in> ?R" by blast+
      then have "(s, t) \<in> ?R" by (auto simp: conversion_def)
      then show ?thesis by blast
    qed
  qed
qed

end

end

end

end

context kb_irun
begin

interpretation lop: SN_order_pair lless lleq by (rule lop)

lemma
  "((s, t) \<in> E i - E\<^sub>\<omega> \<longrightarrow> (s, t) \<in> (\<Union>\<beta>\<in>{\<beta>. ({#s, s, s, t, t, t#}, \<beta>) \<in> lless}. lstep E\<^sub>\<omega> R\<^sub>\<omega> \<beta>)\<^sup>\<leftrightarrow>\<^sup>*) \<and>
   ((s, t) \<in> R i - R\<^sub>\<omega> \<longrightarrow> (s, t) \<in> (\<Union>\<beta>\<in>{\<beta>. ({#s, s, t#}, \<beta>) \<in> lless}. lstep E\<^sub>\<omega> R\<^sub>\<omega> \<beta>)\<^sup>\<leftrightarrow>\<^sup>*)"
  (is "(_ \<longrightarrow> _ \<in> ?E s t) \<and> (_ \<longrightarrow> _ \<in> ?R s t)")
proof (induct "{#s, t#}" arbitrary: s t i rule: SN_induct [OF lop.SN])
  case IH: (1 s t)
  def Est \<equiv> "?E s t"
  let ?\<alpha> = "{#s, s, s, t, t, t#}"
  let ?S = "(relto ({\<cdot>\<rhd>} \<union> {\<succ>}) {\<cdot>\<unrhd>})\<^sup>+"
  show ?case
  proof (intro conjI impI)
    assume *: "(s, t) \<in> E i - E\<^sub>\<omega>"
    then have "\<exists>j>i. (s, t) \<notin> E j" using Suc_le_eq unfolding E\<^sub>\<omega>_def by blast
    moreover def j \<equiv> "LEAST j. j > i \<and> (s, t) \<notin> E j"
    ultimately have "j > i" and not_j: "(s, t) \<notin> E j" by (metis (lifting) LeastI)+
    then have "j - 1 < j" and [simp]: "Suc (j - Suc 0) = j" by auto
    from not_less_Least [OF this(1) [unfolded j_def], folded j_def] and * and \<open>j > i\<close>
      have "(s, t) \<in> E (j - 1)" by (cases "j = Suc i") auto
    with KB_lstep_E_subset [OF _ irun [of "j - 1"], of s t] and run_R_less and not_j
      have "(s, t) \<in> ((\<Union>\<beta>\<in>{\<beta>. (?\<alpha>, \<beta>) \<in> lless}. lstep (E j) (R j) \<beta>)\<^sup>\<leftrightarrow>)\<^sup>*" (is "_ \<in> ?Ej") by (simp add: conversion_def)
    then have "(s, t) \<in> Est"
    proof (induct)
      case (step u v)
      moreover have "(u, v) \<in> Est"
      proof -
        from step(2) obtain \<beta> where \<beta>: "(?\<alpha>, \<beta>) \<in> lless"
          and uv: "(u, v) \<in> (lstep (E j) (R j) \<beta>)\<^sup>\<leftrightarrow>" by blast
        { fix u v assume "(u, v) \<in> lstep (E j) (R j) \<beta>"
          then consider (eq) C l r \<sigma> where "(l, r) \<in> E j" and "u = C\<langle>l \<cdot> \<sigma>\<rangle>" and "v = C\<langle>r \<cdot> \<sigma>\<rangle>"
            and "(\<beta>, {#u, u, u, v, v, v#}) \<in> lleq"
          | (rule) C l r \<sigma> where "(l, r) \<in> R j" and "u = C\<langle>l \<cdot> \<sigma>\<rangle>" and "v = C\<langle>r \<cdot> \<sigma>\<rangle>"
            and "(\<beta>, {#u, l, r#}) \<in> lleq"
            by (auto elim!: lstep_cases rstep_rule.cases)
          then have "(u, v) \<in> Est"
          proof (cases)
            case eq
            then consider "(l, r) \<in> E\<^sub>\<omega>" | "(l, r) \<in> E j - E\<^sub>\<omega>" by blast
            then show ?thesis
            proof (cases)
              case 1
              then have "(u, v) \<in> rstep E\<^sub>\<omega>" using eq by auto
              with eq have "(u, v) \<in> lstep E\<^sub>\<omega> R\<^sub>\<omega> \<beta>" by (auto intro: lstep_E_I)
              then show ?thesis using \<beta> by (auto simp: Est_def)
            next
              case 2
              from \<beta> and eq have "(?\<alpha>, {#u, u, u, v, v, v#}) \<in> lless" (is "(_, ?\<beta>) \<in> _")
                using lop.compat_S_NS_point by blast
              then obtain xs X and ys Y where "?\<alpha> = mset xs + X" and "?\<beta> = mset ys + Y"
                and "length xs = length ys"
                and "X \<noteq> {#}" and "\<forall>i<length ys. xs ! i \<cdot>\<unrhd> ys ! i"
                and "\<forall>y. y \<in># Y \<longrightarrow> (\<exists>x. x \<in># X \<and> (x, y) \<in> ?S)"
                unfolding s_mul_ext_def by blast
              have "({#s, t#}, {#l, r#}) \<in> lless" sorry
              from IH [OF this] and 2
              show ?thesis sorry
            qed
          next
            case rule
            then consider "(l, r) \<in> R\<^sub>\<omega>" | "(l, r) \<in> R j - R\<^sub>\<omega>" by blast
            then show ?thesis
            proof (cases)
              case 1
              with rule have "(u, v) \<in> lstep E\<^sub>\<omega> R\<^sub>\<omega> \<beta>" by (auto intro: lstep_R_I)
              then show ?thesis using \<beta> by (auto simp: Est_def)
            next
              case 2 show ?thesis sorry
            qed
          qed }
        note lstep = this
        show ?thesis using uv by (auto dest!: lstep simp: Est_def conversion_inv)
      qed
      ultimately show ?case by (auto simp: Est_def conversion_def)
    qed (simp add: Est_def)
    then show "(s, t) \<in> ?E s t" by (simp add: Est_def)
  next
    assume "(s, t) \<in> R i - R\<^sub>\<omega>"
    show "(s, t) \<in> ?R s t" sorry
  qed
qed

(*
interpretation lstep: incfam "\<lambda>i. lstep (E i) (R i)" "\<lambda>x y. (y, x) \<in> lless"
apply (standard)
apply (insert lstep_inc wf_lless)
apply (auto simp: SN_iff_wf converse_unfold)
done

lemma lstep_subset:
  "lstep (E i) (R i) a \<subseteq> lstep (E (Suc i)) (R (Suc i)) a \<union> (lstep.r_bel (Suc i) a)\<^sup>\<leftrightarrow>\<^sup>*"
using KB_lstep_subset [OF irun [of i], of a] by (auto simp: lstep.r_bel_def)
*)

(*
lemma
  assumes "E\<^sub>\<omega> = {}"
  shows "lstep.r_per a = lstep E\<^sub>\<omega> R\<^sub>\<omega> a"
proof (intro equalityI subrelI)
  fix s t assume "(s, t) \<in> lstep.r_per a"
  then show "(s, t) \<in> lstep E\<^sub>\<omega> R\<^sub>\<omega> a"
    using assms
    apply (auto simp: lstep.r_per_def E\<^sub>\<omega>_def)
apply (auto simp: lstep.r_per_def)

lemma
  "E\<^sub>\<omega> = {} \<Longrightarrow> lstep.r_per a = lstep {} R\<^sub>\<omega> a"
apply (auto simp: lstep.r_per_def)

(*Lemma 7*)
lemma
  "lstep.r_inf a \<subseteq> (lstep.r_per a)\<^sup>\<leftrightarrow>\<^sup>*"
by (rule lstep.r_inf_r_per_conv [OF lstep_subset])

lemma
  "(lstep (\<Union>i. R i) (\<Union>i. E i) \<alpha>)\<^sup>\<leftrightarrow> \<subseteq> (lstep R\<^sub>\<omega> E\<^sub>\<omega> \<alpha>)\<^sup>\<leftrightarrow>\<^sup>*"

lemma
  assumes nonfail: "E\<^sub>\<omega> = {}"
    and fair: "CP R\<^sub>\<omega> \<subseteq> (\<Union>i. (rstep (E i))\<^sup>\<leftrightarrow>)"
  shows False
oops
*)

end

end
