(*
License: LGPL (see file COPYING.LESSER)
*)
theory CP
imports
  Trs
  Unification
  "$HOME/Repos/permutations/Permutations"
begin

no_notation
  less_multiset (infix "<#" 50)

fun permute_term :: "'v perm \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term" where
  "permute_term p (Var x) = Var (permute_atom p x)" |
  "permute_term p (Fun f ts) = Fun f (map (permute_term p) ts)"

interpretation term_pt: pt permute_term
  by (default) (induct_tac [!] x, auto simp: map_idI)

adhoc_overloading
  PERMUTE permute_term and
  FRESH term_pt.fresh term_pt.fresh_set

interpretation terms_pt: list_pt permute_term ..

adhoc_overloading
  PERMUTE terms_pt.permute_list and
  FRESH terms_pt.fresh

lemma supp_Var:
  fixes x :: "'v :: infinite"
  shows "term_pt.supp (Var x) = {x}"
  using assms and infinite_UNIV
  by (auto simp: term_pt.supp_def swap_atom)

lemma permute_Fun:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (Fun f ss) = Fun f (p \<bullet> ss)"
  by (induct ss) (simp)+

lemma supp_Fun':
  "term_pt.supp (Fun f ss) = pt.supp (op \<bullet>) ss"
  by (simp only: term_pt.supp_def terms_pt.supp_def permute_Fun) auto

lemma supp_Fun:
  "term_pt.supp (Fun f ss) = (\<Union>s\<in>set ss. term_pt.supp s)"
  by (simp add: supp_Fun')

lemma supp_vars_term_eq:
  fixes t :: "('f, 'v :: infinite) term"
  shows "term_pt.supp t = vars_term t"
  using assms by (induct t) (simp add: supp_Var supp_Fun)+

interpretation subst_permute: fun_pt permute_atom permute_term ..

adhoc_overloading
  PERMUTE "subst_permute.permute_fun" and
  FRESH subst_permute.fresh

lemma eqvt_term_apply_subst [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (t \<cdot> \<sigma>) = (p \<bullet> t) \<cdot> (p \<bullet> \<sigma>)"
  by (induct t) (simp add: subst_permute.permute_fun_def)+

fun permute_ctxt :: "'v perm \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt" where
  "permute_ctxt p \<box> = \<box>" |
  "permute_ctxt p (More f ss1 C ss2) =
    More f (map (permute_term p) ss1) (permute_ctxt p C) (map (permute_term p) ss2)"

interpretation ctxt_pt: pt permute_ctxt
  by (default) (induct_tac [!] x, auto simp: map_idI)

adhoc_overloading
  PERMUTE permute_ctxt and
  FRESH ctxt_pt.fresh

lemma supp_Hole:
  "ctxt_pt.supp \<box> = {}"
  by (simp add: ctxt_pt.supp_def)

lemma permute_More:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (More f ss1 C ss2) = More f (p \<bullet> ss1) (p \<bullet> C) (p \<bullet> ss2)"
  by (simp)

lemma supp_More':
  "ctxt_pt.supp (More f ss1 C ss2) =
    pt.supp (op \<bullet>) ss1 \<union> ctxt_pt.supp C \<union>  pt.supp (op \<bullet>) ss2"
  by (simp only: ctxt_pt.supp_def terms_pt.supp_def) auto

lemma supp_More:
  "ctxt_pt.supp (More f ss1 C ss2) =
    (\<Union>s\<in>set ss1. term_pt.supp s) \<union> ctxt_pt.supp C \<union> (\<Union>s\<in>set ss2. term_pt.supp s)"
  by (simp add: supp_More')

lemma supp_vars_ctxt_eq:
  fixes C :: "('f, 'v :: infinite) ctxt"
  shows "ctxt_pt.supp C = vars_ctxt C"
  by (induct C) (auto simp: supp_Hole supp_More supp_vars_term_eq)

lemma eqvt_term_apply_ctxt [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> (C\<langle>t\<rangle>) = (p \<bullet> C)\<langle>p \<bullet> t\<rangle>"
  by (induct C) simp+

interpretation rule_pt: prod_pt permute_term permute_term ..

adhoc_overloading
  PERMUTE "rule_pt.permute_prod" and
  FRESH rule_pt.fresh rule_pt.fresh_set

lemma supp_vars_rule_eq:
  fixes r :: "('f, 'v :: infinite) rule"
  shows "rule_pt.supp r = vars_term (fst r) \<union> vars_term (snd r)"
  by (cases r) (simp add: supp_vars_term_eq)

lemma rstepD [dest]:
  assumes "(s, t) \<in> rstep R"
  shows "\<exists>l r C \<sigma>. (l, r) \<in> R \<and> s = C\<langle>l \<cdot> \<sigma>\<rangle> \<and> t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using rstepE [OF assms] by metis

interpretation trs_pt: set_pt rule_pt.permute_prod ..

adhoc_overloading
  PERMUTE trs_pt.permute_set and
  FRESH trs_pt.fresh

lemma rule_mem_trs_iff [iff]:
  fixes p :: "('v :: infinite) perm"
  shows "(p \<bullet> s, p \<bullet> t) \<in> p \<bullet> R \<longleftrightarrow> (s, t) \<in> R"
  unfolding rule_pt.permute_prod.simps [symmetric] trs_pt.mem_permute_iff ..

lemma inv_rule_mem_trs_simps [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "(-p \<bullet> s, -p \<bullet> t) \<in> R \<longleftrightarrow> (s, t) \<in> p \<bullet> R"
    and "(s, t) \<in> -p \<bullet> R \<longleftrightarrow> (p \<bullet> s, p \<bullet> t) \<in> R"
  unfolding rule_pt.permute_prod.simps [symmetric]
  by (metis trs_pt.inv_mem_simps(1))
     (metis trs_pt.inv_mem_simps(2))

lemma eqvt_rstep [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "p \<bullet> rstep R = rstep (p \<bullet> R)"
proof
  show "rstep (p \<bullet> R) \<subseteq> p \<bullet> rstep R"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep (p \<bullet> R)"
    then show "(s, t) \<in> p \<bullet> rstep R"
    proof (induct)
      case (IH C \<sigma> l r)
      have "(l, r) \<in> p \<bullet> R" by fact
      then have "(p \<bullet> (-p) \<bullet> l, p \<bullet> (-p) \<bullet> r) \<in> p \<bullet> R" by simp+
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> R" by simp
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> rstep R" by blast
      then have "((-p \<bullet> C)\<langle>-p \<bullet> l \<cdot> (-p \<bullet> \<sigma>)\<rangle>, (-p \<bullet> C)\<langle>-p \<bullet> r \<cdot> (-p \<bullet> \<sigma>)\<rangle>) \<in> rstep R" by blast
      then have "(p \<bullet> (-p \<bullet> C)\<langle>-p \<bullet> l \<cdot> (-p \<bullet> \<sigma>)\<rangle>, p \<bullet> (-p \<bullet> C)\<langle>-p \<bullet> r \<cdot> (-p \<bullet> \<sigma>)\<rangle>) \<in> p \<bullet> rstep R"
        unfolding rule_mem_trs_iff .
      then show "(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> p \<bullet> rstep R" by simp
    qed
  qed
next
  show "p \<bullet> rstep R \<subseteq> rstep (p \<bullet> R)"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> p \<bullet> rstep R"
    then have "(-p \<bullet> s, -p \<bullet> t) \<in> -p \<bullet> p \<bullet> rstep R" by simp
    then have "(-p \<bullet> s, -p \<bullet> t) \<in> rstep R" by simp
    then show "(s, t) \<in> rstep (p \<bullet> R)"
    proof (induct x\<equiv>"-p \<bullet> s" y\<equiv>"-p \<bullet> t")
      case (IH C \<sigma> l r)
      then have "(p \<bullet> l, p \<bullet> r) \<in> p \<bullet> R"
        and [simp]: "s = p \<bullet> C\<langle>l \<cdot> \<sigma>\<rangle>" "t = p \<bullet> C\<langle>r \<cdot> \<sigma>\<rangle>" by simp+
      then have "((p \<bullet> C)\<langle>(p \<bullet> l) \<cdot> (p \<bullet> \<sigma>)\<rangle>, (p \<bullet> C)\<langle>(p \<bullet> r) \<cdot> (p \<bullet> \<sigma>)\<rangle>) \<in> rstep (p \<bullet> R)" by blast
      then show "(s, t) \<in> rstep (p \<bullet> R)" by simp
    qed
  qed
qed

lemma term_apply_subst_Var_Rep_perm [simp]:
  "t \<cdot> (Var \<circ> Rep_perm p) = p \<bullet> t"
  by (induct t) (simp add: permute_atom_def)+

text {*The rewrite relation is invariant under variable renamings.*}
lemma rstep_permute [simp]:
  fixes p :: "('v :: infinite) perm"
  shows "rstep (p \<bullet> R) = rstep R"
proof
  show "rstep (p \<bullet> R) \<subseteq> rstep R"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep (p \<bullet> R)"
    then show "(s, t) \<in> rstep R"
    proof (rule rstepE)
      fix l r C \<sigma>
      assume "(l, r) \<in> p \<bullet> R" and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> -p \<bullet> p \<bullet> R" by simp
      then have "(-p \<bullet> l, -p \<bullet> r) \<in> R" by simp
      then have "((-p \<bullet> l) \<cdot> (Var \<circ> Rep_perm p), (-p \<bullet> r) \<cdot> (Var \<circ> (Rep_perm p))) \<in> rstep R" by blast
      then have "(l, r) \<in> rstep R" by simp
      then show "(s, t) \<in> rstep R" by auto
    qed
  qed
next
  show "rstep R \<subseteq> rstep (p \<bullet> R)"
  proof (rule subrelI)
    fix s t
    assume "(s, t) \<in> rstep R"
    then show "(s, t) \<in> rstep (p \<bullet> R)"
    proof (rule rstepE)
      fix l r C \<sigma>
      assume "(l, r) \<in> R" and [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
      then have "(p \<bullet> l, p \<bullet> r) \<in> p \<bullet> R" by simp
      then have "(C\<langle>(p \<bullet> l) \<cdot> (Var \<circ> Rep_perm (-p)) \<cdot> \<sigma>\<rangle>, C\<langle>(p \<bullet> r) \<cdot> (Var \<circ> Rep_perm (-p)) \<cdot> \<sigma>\<rangle>) \<in> rstep (p \<bullet> R)" by blast
      then show "(s, t) \<in> rstep (p \<bullet> R)" by simp
    qed
  qed
qed

lemma rstep_imp_perm_rstep:
  fixes p :: "('v :: infinite) perm"
  shows "(s, t) \<in> rstep R \<Longrightarrow> (p \<bullet> s, p \<bullet> t) \<in> rstep R"
  by (subst (asm) rule_mem_trs_iff [symmetric]) simp

lemma perm_rstep_imp_rstep:
  fixes p :: "('v :: infinite) perm"
  shows "(p \<bullet> s, p \<bullet> t) \<in> rstep R \<Longrightarrow> (s, t) \<in> rstep R"
  by (subst (asm) rule_mem_trs_iff [symmetric, of _ _ _ "-p"]) simp

lemma rstep_permute_iff [iff]:
  fixes p :: "('v :: infinite) perm"
  shows "(p \<bullet> s, p \<bullet> t) \<in> rstep R \<longleftrightarrow> (s, t) \<in> rstep R"
  by (metis perm_rstep_imp_rstep rstep_imp_perm_rstep)

lemma term_apply_subst_Var_Abs_perm:
  "f \<in> perms \<Longrightarrow> t \<cdot> (Var \<circ> f) = Abs_perm f \<bullet> t"
  by (metis Abs_perm_inverse term_apply_subst_Var_Rep_perm)

lemma finite_term_supp: "finite (term_pt.supp t)"
  unfolding supp_vars_term_eq
  by (induct t) simp+

lemma finite_rule_supp: "finite (rule_pt.supp (l, r))"
  by (simp add: finite_term_supp)

interpretation term_fs: fs permute_term
  by (default) (rule finite_term_supp)

interpretation rule_fs: fs rule_pt.permute_prod
  by (default) (auto simp: finite_rule_supp finite_term_supp)

lemma vars_rule_disjoint:
  fixes l r u v :: "('f, 'v::infinite) term"
  shows "\<exists>p. vars_rule (p \<bullet> (l, r)) \<inter> vars_rule (u, v) = {}"
proof -
  from rule_fs.supp_fresh_set obtain p
    where "rule_pt.supp (p \<bullet> (l, r)) \<sharp> (u, v)" by blast
  from rule_pt.fresh_set_disjoint [OF this]
    show ?thesis
    by (auto simp: supp_vars_rule_eq vars_rule_def)
qed

lemma [termination_simp]:
  "x \<in> set xs \<Longrightarrow> size x < list_size size xs"
  by (induct xs) auto

fun funposs :: "('f, 'v) term \<Rightarrow> pos set" where
  "funposs (Var x) = {}" |
  "funposs (Fun f ts) = {\<epsilon>} \<union> (\<Union>i<length ts. {i <# p | p. p \<in> funposs (ts ! i)})"

lemma funposs_imp_poss:
  assumes "p \<in> funposs t"
  shows "p \<in> poss t"
  using assms by (induct t arbitrary: p) auto

fun varposs :: "('f, 'v) term \<Rightarrow> pos set" where
  "varposs (Var x) = {\<epsilon>}" |
  "varposs (Fun f ts) = (\<Union>i<length ts. {i <# p | p. p \<in> varposs (ts ! i)})"

lemma varposs_imp_poss:
  assumes "p \<in> varposs t"
  shows "p \<in> poss t"
  using assms by (induct t arbitrary: p) auto

lemma poss_simps [symmetric, simp]:
  "poss t = funposs t \<union> varposs t"
  "poss t = varposs t \<union> funposs t"
  "funposs t = poss t - varposs t"
  "varposs t = poss t - funposs t"
  by (induct_tac [!] t) auto

(*TODO: move*)
lemma vars_term_subt_at:
  assumes "p \<in> poss t"
  shows "vars_term (t |_ p) \<subseteq> vars_term t"
  using vars_ctxt_pos_term [OF assms] by simp

definition
  overlap ::
    "('f, 'v::infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow>
    ('f, 'v) rule \<Rightarrow> pos \<Rightarrow> ('f, 'v) rule \<Rightarrow> bool"
where
  "overlap R R' r p r' \<longleftrightarrow>
    (\<exists>p. p \<bullet> r \<in> R) \<and>
    (\<exists>p. p \<bullet> r' \<in> R') \<and>
    vars_rule r \<inter> vars_rule r' = {} \<and>
    p \<in> funposs (fst r') \<and>
    (\<exists>\<sigma>::('f, 'v) subst. fst r \<cdot> \<sigma> = (subt_at (fst r') p) \<cdot> \<sigma>) \<and>
    (p = \<epsilon> \<longrightarrow> \<not> (\<exists>p. p \<bullet> r = r'))"

definition
  CP2 :: "('f, 'v::infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) rule set"
where
  "CP2 R R' = {(replace_at (fst r' \<cdot> \<sigma>) p (snd r \<cdot> \<sigma>), snd r' \<cdot> \<sigma>) |
    r p r' \<sigma>. overlap R R' r p r' \<and> \<sigma> = the_mgu (fst r) (subt_at (fst r') p)}"

abbreviation "CP R \<equiv> CP2 R R"

lemma CP2_I:
  assumes "overlap R R' r p r'" and "\<sigma> = the_mgu (fst r) (subt_at (fst r') p)"
  shows "(replace_at (fst r' \<cdot> \<sigma>) p (snd r \<cdot> \<sigma>), snd r' \<cdot> \<sigma>) \<in> CP2 R R'"
  using assms by (force simp: CP2_def)

(*TODO: move*)
lemma pos_prefix_same [simp]:
  "pos_prefix p p = Some \<epsilon>"
  by (induct p) simp_all

definition "pos_diff p q = the (pos_prefix q p)"

lemma prefix_pos_diff [simp]:
  assumes "p \<le> q"
  shows "p <#> pos_diff q p = q"
  using prefix_exists [OF assms] by (auto simp: pos_diff_def)

lemma vars_term_eqvt:
  "\<pi> \<bullet> vars_term t = vars_term (\<pi> \<bullet> t)"
  by (simp add: supp_vars_term_eq [symmetric] eqvt)

lemma is_imgu_imp_is_mgu:
  assumes "is_imgu \<sigma> E"
  shows "is_mgu \<sigma> E"
  using assms by (auto simp: is_imgu_def is_mgu_def)

(*TODO: move*)
lemma coincidence_lemma':
  assumes "vars_term t \<subseteq> V"
  shows "t \<cdot> (\<sigma> |s V) = t \<cdot> \<sigma>"
  using assms by (metis in_mono in_subst_restrict term_subst_eq)

lemma permute_term_subst_apply_term:
  "(\<pi> \<bullet> t) \<cdot> \<sigma> = t \<cdot> (\<sigma> \<circ> Rep_perm \<pi>)"
  by (induct t) (simp_all add: permute_atom_def)

lemma varposs_iff:
  "p \<in> varposs t \<longleftrightarrow> (\<exists>x. p \<in> poss t \<and> t |_ p = Var x)"
  by (induct t arbitrary: p) auto

lemma poss_subst_apply_term:
  assumes "p \<in> poss (t \<cdot> \<sigma>)" and "p \<notin> funposs t"
  obtains q r x where "p = q <#> r" and "q \<in> poss t" and "t |_ q = Var x" and "r \<in> poss (\<sigma> x)"
using assms
proof (induct t arbitrary: p)
  case (Fun f ts)
  then show ?case by (auto) (metis append_PCons nth_mem subt_at.simps(2))
qed simp

lemma replace_at_subst_rsteps:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes *: "\<And>x. (\<sigma> x, \<tau> x) \<in> (rstep R)\<^sup>*"
    and "p \<in> poss t"
    and "t |_ p = Var x"
  shows "(replace_at (t \<cdot> \<sigma>) p (\<tau> x), t \<cdot> \<tau>) \<in> (rstep R)\<^sup>*"
using assms(2-)
proof (induction t arbitrary: p)
  case (Fun f ts)
  then obtain i q where [simp]: "p = i <# q" and i: "i < length ts"
    and q: "q \<in> poss (ts ! i)" and [simp]: "ts ! i |_ q = Var x" by (cases p) auto
  let ?C = "ctxt_of_pos_term q (ts ! i \<cdot> \<sigma>)"
  let ?ts = "map (\<lambda>t. t \<cdot> \<tau>) ts"
  let ?ss = "take i (map (\<lambda>t. t \<cdot> \<sigma>) ts) @ ?C\<langle>\<tau> x\<rangle> # drop (Suc i) (map (\<lambda>t. t \<cdot> \<sigma>) ts)"
  have "\<forall>j<length ts. (?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*"
  proof (intro allI impI)
    fix j
    assume j: "j < length ts"
    moreover
    { assume [simp]: "j = i"
      have "?ss ! j = ?C\<langle>\<tau> x\<rangle>" using i by (simp add: nth_append_take)
      with Fun.IH [of "ts ! i" q]
      have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" using q and i by simp }
    moreover
    { assume "j < i"
      with i have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_take_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by (auto simp: * subst_rsteps_imp_rsteps) }
    moreover
    { assume "j > i"
      with i and j have "?ss ! j = ts ! j \<cdot> \<sigma>"
        and "?ts ! j = ts ! j \<cdot> \<tau>" by (simp_all add: nth_append_drop_is_nth_conv)
      then have "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by (auto simp: * subst_rsteps_imp_rsteps) }
    ultimately show "(?ss ! j, ?ts ! j) \<in> (rstep R)\<^sup>*" by arith
  qed
  moreover have "i < length ts" by fact
  ultimately show ?case
    by (simp add: args_rsteps_imp_rsteps)
qed simp

(*TODO: move*)
lemma replace_at_ident [simp]:
  assumes "p \<in> poss t" and "t |_ p = s"
  shows "replace_at t p s = t"
  using assms by (metis ctxt_supt_id)

text {*Peaks are either joinable or stem from a critical pair.*}
(*Lemma 5.1.9*)
lemma peak_imp_join_or_CP:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes variable_condition: "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and peak: "(t, u) \<in> (rstep R)\<inverse> O rstep R"
  shows "(t, u) \<in> (rstep R)\<^sup>\<down> \<or> (t, u) \<in> (rstep (CP R))\<^sup>\<leftrightarrow>"
proof -
  from peak obtain s where "(s, t) \<in> rstep R" and "(s, u) \<in> rstep R" by blast
  then obtain p\<^sub>1 l\<^sub>1' r\<^sub>1' \<sigma>\<^sub>1' p\<^sub>2 l\<^sub>2 r\<^sub>2 \<sigma>\<^sub>2
    where rule1': "(l\<^sub>1', r\<^sub>1') \<in> R" and rule2: "(l\<^sub>2, r\<^sub>2) \<in> R"
    and p1: "p\<^sub>1 \<in> poss s" and p2: "p\<^sub>2 \<in> poss s"
    and s': "replace_at s p\<^sub>1 (l\<^sub>1' \<cdot> \<sigma>\<^sub>1') = s" and t': "t = replace_at s p\<^sub>1 (r\<^sub>1' \<cdot> \<sigma>\<^sub>1')"
    and s2: "replace_at s p\<^sub>2 (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) = s" and u: "u = replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)"
    by (auto simp: rstep_iff_rstep_r_p_s rstep_r_p_s_def Let_def)
  have sp2: "s |_ p\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using s2 and p2 by (metis replace_at_subt_at)
  
  from vars_rule_disjoint obtain \<pi>
    where *: "vars_rule (\<pi> \<bullet> (l\<^sub>1', r\<^sub>1')) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}" ..
  def l\<^sub>1 \<equiv> "\<pi> \<bullet> l\<^sub>1'" and r\<^sub>1 \<equiv> "\<pi> \<bullet> r\<^sub>1'" and \<sigma>\<^sub>1 \<equiv> "(Var \<circ> Rep_perm (-\<pi>)) \<circ>\<^sub>s \<sigma>\<^sub>1'"

  have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_rule (l\<^sub>2, r\<^sub>2) = {}" using * by (simp add: l\<^sub>1_def r\<^sub>1_def)
  have s: "replace_at s p\<^sub>1 (l\<^sub>1 \<cdot> \<sigma>\<^sub>1) = s" and t: "t = replace_at s p\<^sub>1 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1)"
    by (simp_all add: s' t' l\<^sub>1_def r\<^sub>1_def \<sigma>\<^sub>1_def)
  have sp1: "s |_ p\<^sub>1 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using s and p1 by (metis replace_at_subt_at)

  show ?thesis
  proof (cases "p\<^sub>1 \<bottom> p\<^sub>2")
    assume "p\<^sub>1 \<bottom> p\<^sub>2"
    moreover
    have "(s, t) \<in> rstep_r_p_s R (l\<^sub>1', r\<^sub>1') p\<^sub>1 \<sigma>\<^sub>1'"
      by (simp add: p1 rstep_r_p_s_def Let_def s' t' rule1')
    moreover
    have "(s, u) \<in> rstep_r_p_s R (l\<^sub>2, r\<^sub>2) p\<^sub>2 \<sigma>\<^sub>2"
      by (simp add: p2 rstep_r_p_s_def Let_def s2 u rule2)
    ultimately
    obtain v where "(t, v) \<in> rstep R" and "(u, v) \<in> rstep R"
      by (blast dest: parallel_steps rstep_r_p_s_imp_rstep)
    then show ?thesis by blast
  next
    assume "\<not> (p\<^sub>1 \<bottom> p\<^sub>2)"
    then have "p\<^sub>1 \<le> p\<^sub>2 \<or> p\<^sub>2 \<le> p\<^sub>1" by (metis parallel_pos)
    then show ?thesis
    proof
      def p \<equiv> "pos_diff p\<^sub>1 p\<^sub>2"
      assume "p\<^sub>2 \<le> p\<^sub>1"
      then have *:"p\<^sub>1 = p\<^sub>2 <#> p" by (simp add: p_def)
      then have **: "t = replace_at s p\<^sub>2 (replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1))"
        by (simp add: t ctxt_of_pos_term_append p2 sp2)
      show ?thesis
      proof (cases "p \<in> funposs l\<^sub>2 \<and> (p = \<epsilon> \<longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2)))")
        case True
        def \<sigma>' \<equiv> "\<lambda>x. if x \<in> vars_rule (l\<^sub>1, r\<^sub>1) then \<sigma>\<^sub>1 x else \<sigma>\<^sub>2 x"
        have "-\<pi> \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" and "0 \<bullet> (l\<^sub>2, r\<^sub>2): R"
          using rule1' and rule2 by (simp_all add: l\<^sub>1_def r\<^sub>1_def)
        then have "\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) \<in> R" and "\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>2, r\<^sub>2) \<in> R" by blast+
        moreover
        from True have p: "p \<in> funposs l\<^sub>2" and "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2))" by auto
        moreover
        have unif: "l\<^sub>1 \<cdot> \<sigma>' = (l\<^sub>2 |_ p) \<cdot> \<sigma>'"
        proof -
          note coinc = coincidence_lemma' [of l\<^sub>1 "vars_rule (l\<^sub>1, r\<^sub>1)"]
          have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term (l\<^sub>2 |_ p) = {}"
            using vars_term_subt_at [OF funposs_imp_poss [OF p]] and disj
            by (auto simp: vars_rule_def)
          have "l\<^sub>1 \<cdot> \<sigma>' = l\<^sub>1 \<cdot> (\<sigma>' |s vars_rule (l\<^sub>1, r\<^sub>1))"
            using coinc by (simp add: vars_rule_def)
          also have "\<dots> = l\<^sub>1 \<cdot> (\<sigma>\<^sub>1 |s vars_rule (l\<^sub>1, r\<^sub>1))" by (simp add: \<sigma>'_def)
          also have "\<dots> = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using coinc by (simp add: vars_rule_def)
          also have "\<dots> = (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) |_ p" using subt_at_append [OF p2] and sp1 by (simp add: * sp2)
          also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> \<sigma>\<^sub>2" using funposs_imp_poss [OF p] by simp
          also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> (\<sigma>\<^sub>2 |s vars_term (l\<^sub>2 |_ p))"
            by (simp add: coincidence_lemma [symmetric])
          also have "\<dots> = (l\<^sub>2 |_ p) \<cdot> (\<sigma>' |s vars_term (l\<^sub>2 |_ p))" using disj by (simp add: \<sigma>'_def)
          finally show ?thesis by (simp add: coincidence_lemma [symmetric])
        qed
        ultimately have overlap: "overlap R R (l\<^sub>1, r\<^sub>1) p (l\<^sub>2, r\<^sub>2)"
          using disj by (auto simp: overlap_def)

        def \<sigma> \<equiv> "the_mgu l\<^sub>1 (l\<^sub>2 |_ p)"
        have CP: "(replace_at (l\<^sub>2 \<cdot> \<sigma>) p (r\<^sub>1 \<cdot> \<sigma>), r\<^sub>2 \<cdot> \<sigma>) \<in> CP R"
          using CP2_I [OF overlap refl] by (simp add: \<sigma>_def)

        from the_mgu [OF unif] have "is_mgu \<sigma> {(l\<^sub>1, l\<^sub>2 |_ p)}"
          by (auto simp: \<sigma>_def is_mgu_def unifiers_def) (metis the_mgu)
        with unif obtain \<tau> where \<sigma>': "\<sigma>' = \<sigma> \<circ>\<^sub>s \<tau>" by (auto simp: is_mgu_def unifiers_def)
        have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1), r\<^sub>2 \<cdot> \<sigma>\<^sub>2) \<in> rstep (CP R)"
        proof -
          have "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>'"
          proof -
            have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term l\<^sub>2 = {}" using disj by (auto simp: vars_rule_def)
            have "l\<^sub>2 \<cdot> \<sigma>' = l\<^sub>2 \<cdot> (\<sigma>' |s vars_term l\<^sub>2)" by (rule coincidence_lemma)
            also have "\<dots> = l\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_term l\<^sub>2)" using disj by (simp add: \<sigma>'_def)
            also have "\<dots> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: coincidence_lemma [symmetric])
            finally show ?thesis ..
          qed
          moreover
          have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = r\<^sub>1 \<cdot> \<sigma>'"
          proof -
            note coinc = coincidence_lemma' [of r\<^sub>1 "vars_rule (l\<^sub>1, r\<^sub>1)"]
            have "r\<^sub>1 \<cdot> \<sigma>' = r\<^sub>1 \<cdot> (\<sigma>' |s vars_rule (l\<^sub>1, r\<^sub>1))"
              using coinc by (simp add: vars_rule_def)
            also have "\<dots> = r\<^sub>1 \<cdot> (\<sigma>\<^sub>1 |s vars_rule (l\<^sub>1, r\<^sub>1))" by (simp add: \<sigma>'_def)
            also have "\<dots> = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" using coinc by (simp add: vars_rule_def)
            finally show ?thesis ..
          qed
          moreover
          have "r\<^sub>2 \<cdot> \<sigma>\<^sub>2 = r\<^sub>2 \<cdot> \<sigma>'"
          proof -
            have disj: "vars_rule (l\<^sub>1, r\<^sub>1) \<inter> vars_term r\<^sub>2 = {}" using disj by (auto simp: vars_rule_def)
            have "r\<^sub>2 \<cdot> \<sigma>' = r\<^sub>2 \<cdot> (\<sigma>' |s vars_term r\<^sub>2)" by (rule coincidence_lemma)
            also have "\<dots> = r\<^sub>2 \<cdot> (\<sigma>\<^sub>2 |s vars_term r\<^sub>2)" using disj by (simp add: \<sigma>'_def)
            also have "\<dots> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2" by (simp add: coincidence_lemma [symmetric])
            finally show ?thesis ..
          qed
          ultimately
          show ?thesis
            using rstep.rule [OF CP, THEN rstep.subst, of \<tau>]
            by (simp add: \<sigma>' ctxt_of_pos_term_subst [OF funposs_imp_poss [OF p, THEN poss_imp_subst_poss]])
        qed
        then have "(t, u) \<in> rstep (CP R)" by (auto simp: ** u)
        then show ?thesis by blast
      next
        case False
        then have "(p = \<epsilon> \<and> (\<exists>\<pi>. \<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2))) \<or> p \<notin> funposs l\<^sub>2" by blast
        then show ?thesis
        proof (elim disjE conjE exE)
          fix \<pi>
          assume [simp]: "p = \<epsilon>" and "\<pi> \<bullet> (l\<^sub>1, r\<^sub>1) = (l\<^sub>2, r\<^sub>2)"
          then have "p\<^sub>2 = p\<^sub>1" and "l\<^sub>2 = \<pi> \<bullet> l\<^sub>1" and [simp]: "r\<^sub>2 = \<pi> \<bullet> r\<^sub>1" by (simp_all add: *)
          with sp1 and sp2
            have "l\<^sub>1 \<cdot> \<sigma>\<^sub>1 = l\<^sub>1 \<cdot> (\<sigma>\<^sub>2 \<circ> Rep_perm \<pi>)" by (simp add: permute_term_subst_apply_term)
          moreover
          have "vars_term r\<^sub>1 \<subseteq> vars_term l\<^sub>1"
            using variable_condition [THEN bspec, OF rule1']
            and atom_set_pt.permute_set_subset [of "vars_term r\<^sub>1'" "vars_term l\<^sub>1'"]
            by (auto simp add: l\<^sub>1_def r\<^sub>1_def supp_vars_term_eq [symmetric] term_fs.supp_eqvt [symmetric])
          ultimately
          have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = r\<^sub>1 \<cdot> (\<sigma>\<^sub>2 \<circ> Rep_perm \<pi>)" using vars_term_subset_subst_eq by blast
          then have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = \<pi> \<bullet> r\<^sub>1 \<cdot> \<sigma>\<^sub>2" by (simp add: permute_term_subst_apply_term)
          then show ?thesis by (auto simp: ** u)
        next
          have 1: "l\<^sub>2 \<cdot> \<sigma>\<^sub>2 |_ p = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" by (metis "*" p2 sp1 sp2 subt_at_append)
          have "p \<in> poss (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)" by (metis "*" p1 poss_append_poss sp2)
          moreover
          assume "p \<notin> funposs l\<^sub>2"
          ultimately obtain q\<^sub>1 q\<^sub>2 x
            where p: "p = q\<^sub>1 <#> q\<^sub>2" and q1: "q\<^sub>1 \<in> poss l\<^sub>2"
            and l2q1 [simp]: "l\<^sub>2 |_ q\<^sub>1 = Var x" and q2: "q\<^sub>2 \<in> poss (\<sigma>\<^sub>2 x)"
            by (rule poss_subst_apply_term)
          then have 2: "\<sigma>\<^sub>2 x |_ q\<^sub>2 = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" using 1 by simp

          def \<sigma>\<^sub>2' \<equiv> "\<lambda>y. if y = x then replace_at (\<sigma>\<^sub>2 x) q\<^sub>2 (r\<^sub>1 \<cdot> \<sigma>\<^sub>1) else \<sigma>\<^sub>2 y"

          have x: "(\<sigma>\<^sub>2 x, \<sigma>\<^sub>2' x) \<in> rstep R"
          proof -
            let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma>\<^sub>2 x)"
            have "(l\<^sub>1, r\<^sub>1) \<in> rstep R"
              using rstep.rule [OF rule1'] by (simp add: l\<^sub>1_def r\<^sub>1_def)
            then have "(l\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<sigma>\<^sub>1) \<in> rstep R" by blast
            then have "(?C\<langle>l\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>, ?C\<langle>r\<^sub>1 \<cdot> \<sigma>\<^sub>1\<rangle>) \<in> rstep R" by blast
            then show ?thesis using 2 and q2 by (simp add: \<sigma>\<^sub>2'_def)
          qed

          have r2: "(r\<^sub>2 \<cdot> \<sigma>\<^sub>2, r\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> (rstep R)\<^sup>*"
          proof (induct r\<^sub>2)
            case (Var y)
            show ?case using x by (cases "y = x") (simp_all add: \<sigma>\<^sub>2'_def)
          next
            case (Fun f ts)
            then have "\<forall>i<length (map (\<lambda>t. t \<cdot> \<sigma>\<^sub>2) ts).
              (map (\<lambda>t. t \<cdot> \<sigma>\<^sub>2 ) ts ! i, map (\<lambda>t. t \<cdot> \<sigma>\<^sub>2') ts ! i) \<in> (rstep R)\<^sup>*" by auto
            from args_rsteps_imp_rsteps [OF _ this] show ?case by simp
          qed
          have "(u, replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2')) \<in> (rstep R)\<^sup>*" by (auto simp: r2 u rsteps_closed_ctxt)
          moreover
          have "(t, replace_at s p\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2')) \<in> (rstep R)\<^sup>*"
          proof -
            have "replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1) = replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) q\<^sub>1 (\<sigma>\<^sub>2' x)"
              using 1 and 2 and q1 and q2
              apply (simp add: \<sigma>\<^sub>2'_def)
              by (metis `l\<^sub>2 |_ q\<^sub>1 = Var x` `p = q\<^sub>1 <#> q\<^sub>2` `p \<in> poss (l\<^sub>2 \<cdot> \<sigma>\<^sub>2)` ctxt_ctxt ctxt_of_pos_term_append poss_append_poss subst_apply_term.simps(1) subt_at_subst)
            moreover
            have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) q\<^sub>1 (\<sigma>\<^sub>2' x), l\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> (rstep R)\<^sup>*"
            proof -
              have "\<And>y. (\<sigma>\<^sub>2 y, \<sigma>\<^sub>2' y) \<in> (rstep R)\<^sup>*"
                using x by (case_tac "y = x") (simp_all add: \<sigma>\<^sub>2'_def)
              from replace_at_subst_rsteps [OF this q1 l2q1] show ?thesis .
            qed
            moreover
            have "(l\<^sub>2 \<cdot> \<sigma>\<^sub>2', r\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> rstep R" using rule2 by blast
            ultimately
            have "(replace_at (l\<^sub>2 \<cdot> \<sigma>\<^sub>2) p (r\<^sub>1 \<cdot> \<sigma>\<^sub>1), r\<^sub>2 \<cdot> \<sigma>\<^sub>2') \<in> (rstep R)\<^sup>*" by auto
            then show ?thesis by (auto simp: p2 ** rsteps_closed_ctxt)
          qed
          ultimately show ?thesis by blast
        qed
      qed
    next
      assume "p\<^sub>1 \<le> p\<^sub>2"
      show ?thesis sorry
    qed
  qed
qed

lemma join_ctxt [intro]:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms obtain u where "(s, u) \<in> (rstep R)\<^sup>*" and "(t, u) \<in> (rstep R)\<^sup>*" by auto
  then have "(C\<langle>s\<rangle>, C\<langle>u\<rangle>) \<in> (rstep R)\<^sup>*" and "(C\<langle>t\<rangle>, C\<langle>u\<rangle>) \<in> (rstep R)\<^sup>*" by (auto intro: rsteps_closed_ctxt)
  then show ?thesis by blast
qed

lemma join_subst [intro]:
  assumes "(s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms obtain u where "(s, u) \<in> (rstep R)\<^sup>*" and "(t, u) \<in> (rstep R)\<^sup>*" by auto
  then have "(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*" and "(t \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (rstep R)\<^sup>*" by (auto intro: rsteps_closed_subst)
  then show ?thesis by blast
qed

lemma join_sym:
  "(s, t) \<in> A\<^sup>\<down> \<Longrightarrow> (t, s) \<in> A\<^sup>\<down>"
  by auto

lemma CP_join_imp_WCR:
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
    and *: "\<forall>(s, t)\<in>CP R. (s, t) \<in> (rstep R)\<^sup>\<down>"
  shows "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down>"
proof -
  have "(rstep (CP R))\<^sup>\<leftrightarrow> \<subseteq> (rstep R)\<^sup>\<down>"
  proof (intro subrelI)
    fix s t
    assume "(s, t) \<in> (rstep (CP R))\<^sup>\<leftrightarrow>"
    then have "(s, t) \<in> rstep (CP R) \<or> (t, s) \<in> rstep (CP R)" by auto
    then show "(s, t) \<in> (rstep R)\<^sup>\<down>"
    proof
      assume "(s, t) \<in> rstep (CP R)"
      then obtain C \<sigma> u v where cp: "(u, v) \<in> CP R"
        and [simp]: "s = C\<langle>u \<cdot> \<sigma>\<rangle>" "t = C\<langle>v \<cdot> \<sigma>\<rangle>" by auto
      from * [THEN bspec, OF cp] show ?thesis by auto
    next
      assume "(t, s) \<in> rstep (CP R)"
      then obtain C \<sigma> u v where cp: "(u, v) \<in> CP R"
        and [simp]: "t = C\<langle>u \<cdot> \<sigma>\<rangle>" "s = C\<langle>v \<cdot> \<sigma>\<rangle>" by auto
      show ?thesis by (rule join_sym) (insert * [THEN bspec, OF cp], auto)
    qed
  qed
  with peak_imp_join_or_CP [OF assms(1)] show ?thesis by blast
qed

lemma CP_imp_peak:
  fixes R :: "('f, 'v :: infinite) trs"
  assumes "(s, t) \<in> CP2 R S"
  shows "(s, t) \<in> (rstep R)\<inverse> O (rstep S)"
proof -
  from assms [unfolded CP2_def] obtain r p r' \<sigma>
    where lap: "overlap R S r p r'"
    and "\<sigma> = the_mgu (fst r) (fst r' |_ p)"
    and [simp]: "s = replace_at (fst r' \<cdot> \<sigma>) p (snd r \<cdot> \<sigma>)"
                "t = snd r' \<cdot> \<sigma>"
    by auto
  from lap obtain \<pi>\<^sub>1 \<pi>\<^sub>2 and \<sigma> :: "('f, 'v) subst"
    where "\<pi>\<^sub>1 \<bullet> r \<in> R" and "\<pi>\<^sub>2 \<bullet> r' \<in> S"
    and "fst r \<cdot> \<sigma> = (subt_at (fst r') p) \<cdot> \<sigma>"
    by (auto simp: overlap_def)
  then have "\<pi>\<^sub>1 \<bullet> r \<in> rstep R" and "\<pi>\<^sub>2 \<bullet> r' \<in> rstep S"
    by (metis rstep.rule surjective_pairing)+
  then have "r \<in> rstep R" and "r' \<in> rstep S"
    by (metis perm_rstep_imp_rstep rule_pt.permute_prod.simps surjective_pairing)+
  oops

lemma WCR_imp_CP_join:
  assumes "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down>"
   and "(s, t) \<in> CP R"
  shows "(s, t) \<in> (rstep R)\<^sup>\<down>"
proof -
  from assms(2) [unfolded CP2_def]
  oops

text {*The Critical Pair Lemma*}
lemma CP:
  assumes "\<forall>(l, r)\<in>R. vars_term r \<subseteq> vars_term l"
  shows "(rstep R)\<inverse> O (rstep R) \<subseteq> (rstep R)\<^sup>\<down> \<longleftrightarrow> (\<forall>(s, t)\<in>CP R. (s, t) \<in> (rstep R)\<^sup>\<down>)"
(*
  using CP_join_imp_WCR [OF assms] and WCR_imp_CP_join [of R] by auto
*)
oops

end

