(*
License: LGPL (see file COPYING.LESSER)
*)
theory Knuth_Bendix_Order
imports
  Main
  "~~/src/HOL/Library/Sublist"
  Simple_Termination
begin

fun lexteq :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool"
where
  "lexteq P [] [] \<longleftrightarrow> True" |
  "lexteq P (x#xs) [] \<longleftrightarrow> True" |
  "lexteq P [] (y#ys) \<longleftrightarrow> False" |
  "lexteq P (x#xs) (y#ys) \<longleftrightarrow> P x y \<or> (x = y \<and> lexteq P xs ys)"

fun lext :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool"
where
  "lext P [] [] \<longleftrightarrow> False" |
  "lext P (x#xs) [] \<longleftrightarrow> True" |
  "lext P [] (y#ys) \<longleftrightarrow> False" |
  "lext P (x#xs) (y#ys) \<longleftrightarrow> P x y \<or> (x = y) \<and> lext P xs ys"

lemmas lex_induct = lexteq.induct

lemma lext_imp_lexteq:
  assumes "lext P xs ys"
  shows "lexteq P xs ys"
  using assms by (induct P xs ys rule: lex_induct) auto

lemma lext_mono:
  assumes "\<And>x y. P x y \<longrightarrow> Q x y"
    and "lext P xs ys"
  shows "lext Q xs ys"
  using assms by (induct P\<equiv>P xs ys rule: lex_induct) auto

lemma lexteq_mono:
  assumes "\<And>x y. P x y \<longrightarrow> Q x y"
    and "lexteq P xs ys"
  shows "lexteq Q xs ys"
  using assms by (induct P\<equiv>P xs ys rule: lex_induct) auto

lemma lexteq_Nil2 [simp]:
  "lexteq P xs []"
  by (induct xs) simp_all

lemma list_emb_rev [simp, intro]:
  assumes "list_emb P xs ys"
  shows "list_emb P (rev xs) (rev ys)"
  using assms by (induct) (auto intro: list_emb_append_mono)

lemma list_emb_rev_iff:
  "list_emb P xs ys \<longleftrightarrow> list_emb P (rev xs) (rev ys)"
proof
  note * =
    list_emb_rev [THEN list_emb_prefix [of _ "rev (rev xs)"], unfolded rev_rev_ident]
    list_emb_Cons2 [THEN list_emb_rev, unfolded rev.simps]
  assume "list_emb P (rev xs) (rev ys)"
  then show "list_emb P xs ys" by (cases _ "rev xs" "rev ys") (auto intro: *)
qed (rule list_emb_rev)

lemma list_emb_rev_induct [consumes 1]:
  assumes "list_emb P xs ys"
    and "\<And>ys. Q [] ys"
    and "\<And>xs ys y. list_emb P xs ys \<Longrightarrow> Q xs ys \<Longrightarrow> Q xs (ys @ [y])"
    and "\<And>x y xs ys. P x y \<Longrightarrow> list_emb P xs ys \<Longrightarrow> Q xs ys \<Longrightarrow> Q (xs @ [x]) (ys @ [y])"
  shows "Q xs ys"
proof -
  have "list_emb P (rev xs) (rev ys)" by (simp add: assms)
  then show ?thesis
  proof (induct _ "rev xs" "rev ys" arbitrary: xs ys)
    case (list_emb_Nil xs)
    then show ?case by (simp add: assms)
  next
    case (list_emb_Cons us u vs)
    then have [simp]: "ys = rev us @ [u]" by (metis rev_eq_Cons_iff)
    then have "Q (rev (rev vs)) (rev us)" using list_emb_Cons by simp
    have "Q (rev (rev vs)) (rev us @ [u])" by (intro assms list_emb_rev) fact+
    then show ?case by simp
  next
    case (list_emb_Cons2 u v us vs)
    then have [simp]: "xs = rev us @ [u]" "ys = rev vs @ [v]" by (metis rev_eq_Cons_iff)+
    show ?case using list_emb_Cons2 by (auto intro: assms)
  qed
qed


inductive list_emb' for P :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
where
  [intro, simp]: "list_emb' P [] ys" |
  [intro]: "list_emb' P xs ys \<Longrightarrow> list_emb' P xs (ys @ [y])" |
  [intro]: "P x y \<Longrightarrow> list_emb' P xs ys \<Longrightarrow> list_emb' P (x # xs) (y # ys)"

lemma list_emb'_imp_list_emb:
  assumes "list_emb' P xs ys"
  shows "list_emb P xs ys"
  using assms by (induct) auto

lemma list_emb'_append2 [intro]:
  assumes "list_emb' P xs ys" shows "list_emb' P xs (ys @ zs)"
  using assms by (induct zs rule: rev_induct) (auto dest: list_emb'.intros(2))

lemma list_emb'_suffix [intro]:
  assumes "list_emb' P xs ys" shows "list_emb' P xs (zs @ ys)"
  using assms
  apply (induct arbitrary: zs)
  apply auto
  apply (metis append_assoc list_emb'.simps)

lemma list_emb'_append_mono:
  "\<lbrakk> list_emb' P ys ys'; list_emb' P xs xs' \<rbrakk> \<Longrightarrow> list_emb' P (xs@ys) (xs'@ys')"
  apply (induct rule: list_emb'.induct)
  apply (metis append_Nil2 list_emb'_append2)
  apply (metis append_assoc list_emb'.simps)

lemma list_emb_imp_list_emb':
  assumes "list_emb P xs ys"
  shows "list_emb' P xs ys"
  using assms apply (induct)
  apply (auto)
  apply (metis append_Cons append_Nil list_emb'_suffix)
  apply (drule list_emb'.intros(3) [of P, OF _ list_emb'.intros(1)[of _ "[]"]])
  apply simp
  by (metis append_Cons append_Nil list_emb'_append_mono)

lemma list_emb_conv: "list_emb P xs ys = list_emb' P xs ys"
  by (metis list_emb'_imp_list_emb list_emb_imp_list_emb')

lemma lexteq_append [simp, intro]:
  assumes "lexteq P xs ys"
  shows "lexteq P (xs@zs) (ys)"
  using assms
  by (induct P xs ys rule: lex_induct) auto


lemma
  assumes "list_emb P\<inverse>\<inverse> ys xs"
  shows "lexteq P xs ys"
  using assms
  apply (induct)
  apply simp_all
  using assms apply (induct rule: list_emb_rev_induct)
  apply simp_all
  oops

end

