(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2011-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Gcd
imports "../TermFun"
begin

(* A TRS definition for mod. Feel free to choose another, if
this simplifies the problem below. *)
fun mod' :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "mod' k n = (if n = 0 then k else
               if k < n then k else mod' (k - n) n)"

define_trs mod'.simps

text \<open>The attribute @{text "trs_subst"} tells the system that @{const "mod'"}
should be used instead of @{const "mod"} for generating a term representation. Of course,
we also have to prove formally that those two constants are equivalent.\<close>
lemma mod_subst[trs_subst]: "x mod y = mod' x y"
proof -
  let ?P = "\<lambda> (x,y). x mod y = mod' x y"
  have "?P (x,y)"
  proof (induct rule: wf_induct[of _ ?P, OF wf_measures[of "[\<lambda> (x,y). x]"]])
    case (1 xy)
    show ?case
    proof (cases xy)
      case (Pair x y)
      have "x mod y = mod' x y" 
      proof (cases "y = 0")
        case True
        thus ?thesis by simp
      next
        case False note oFalse = this
        show ?thesis 
        proof (cases "x < y")
          case True
          with False show ?thesis by simp
        next
          case False
          with oFalse have "x \<noteq> 0" by simp
          with oFalse have smaller: "x - y < x" by simp
          from 1[THEN spec, of "(x-y,y)", unfolded Pair] 
          have ind: "(x-y) mod y = mod' (x-y) y" by simp
          have "mod' x y = mod' (x - y) y" by (simp add: False oFalse)
          also have "\<dots> = (x-y) mod y" using ind by simp
          finally have id: "mod' x y = (x - y) mod y" .
          show ?thesis unfolding id unfolding mod_if[of x y] using False by simp
        qed
      qed
      thus ?thesis unfolding Pair by simp
    qed
  qed
  thus ?thesis by simp
qed

(* taken from HOL/Nitpick.thy *)
function (sequential) gcd :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "gcd x 0 = x"
| "gcd x y = gcd y (x mod y)"
  by pat_completeness auto

define_trs (no_sim) gcd.psimps

termination gcd
  apply (trs_reduce gcd)
  apply expand_trs_problem
  export_termination_problem "gcd.xml"
oops (* no proof yet *)

end
